/* TCMixer 2.0 -- X11 mixer
 * Copyright (C) 1999, 2000 timecop@japan.co.jp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <sys/soundcard.h>

#include "main.xpm"

#define offset (100 - (((y - 36) * 100) / 40))
#define eprint(fmt, arg...) fprintf(stderr, fmt, ##arg)

/* proto */
void draw_switches(int, int);
void draw_slider(int, int);
void set_record(void);
void keyEvent(unsigned int keycode, unsigned int state);
void motionEvent(XMotionEvent *);
void pressEvent(XButtonEvent *);
void releaseEvent(XButtonEvent *);
void setmix(int channel);

/* globals. too many of them, oh well */
int recsource;
int screen;
int depth;
char *dispname;
Window rootwin;
Window win;
Display *display;
GC maingc;
Pixmap mainpanel;
Pixmap mixer;
XImage *handle;
XImage *onhandle;
int mixer_fd;
int dragging = 0;
int currslider = 0;

int mix_type[8] =
    { SOUND_MIXER_VOLUME, SOUND_MIXER_BASS, SOUND_MIXER_TREBLE,
    SOUND_MIXER_PCM, SOUND_MIXER_SYNTH, SOUND_MIXER_CD,
    SOUND_MIXER_LINE, SOUND_MIXER_MIC
};

typedef struct {
    int volume;
    int oldvolume;
    Bool mute;
    Bool cap;
    int type;
    char name[32];
} Mixer;
Mixer mixer_data[8];

/* main program starts here */
void mixer_init(char *mixer_device)
{
    int i, j;
    int devmask;
    char *names[] = SOUND_DEVICE_NAMES;

    mixer_fd = open(mixer_device, O_RDWR, 0);
    if (mixer_fd < 0) {
	perror("Error opening mixer device");
	exit(-1);
    }

    /* read in device capability mask */
    if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &devmask) == -1) {
	eprint("Cannot read device mask\n");
	exit(-1);
    }

    /* read in initial values for the supported channels, set capability
     * mask and channel type. */
    printf("[ ");
    for (i = 0; i < 8; i++) {
	int foo;
	mixer_data[i].type = -1;
	for (j = 0; j < SOUND_MIXER_NRDEVICES; j++) {
	    if (((1 << j) & devmask) & (1 << mix_type[i])) {
		mixer_data[i].type = mix_type[i];
		mixer_data[i].cap = True;
		strcpy(mixer_data[i].name, names[j]);
		ioctl(mixer_fd, MIXER_READ(j), &foo);
		mixer_data[i].volume = ((foo >> 8) + (foo & 0xFF)) / 2;
		mixer_data[i].oldvolume = 0;
		if (mixer_data[i].volume == 0)
		    mixer_data[i].mute = True;
		else
		    mixer_data[i].mute = False;
		printf("%s ", names[j]);
	    }
	}
	if (mixer_data[i].type == -1) {
	    printf("none ");
	    mixer_data[i].cap = False;
	}
    }
    printf("]\n");

    /* read in recording source.  For now, just quit if read is not
     * possible.  In the future, disable selecting record source (2.0 maybe) */
    if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &recsource) == -1) {
	eprint("Error reading mixer recording source\n");
	exit(-1);
    }
}

void make_caps(void)
{
    int i;

    for (i = 0; i < 8; i++) {
	if (mixer_data[i].cap == False) {
	    XCopyArea(display, mainpanel, mixer, maingc, 77, 118, 24, 11,
		      (i * 27) + 8, 18);
	    XCopyArea(display, mainpanel, mixer, maingc, 102, 118, 15, 54,
		      (i * 27) + 11, 29);
	}
    }
}

int main(int argc, char **argv)
{
    static int modify_counter;
    struct mixer_info mixer_info;
    char *mixer_device = NULL;
    int i, j, ret;
    int ch;
    Bool ign = False;
    XEvent Event;
    XEvent bob;
    XpmAttributes xpma;
    XSizeHints sizehints;
    XSetWindowAttributes xattr;

    /* initialize mixer, read in values.  if we fail here, we exit(-1)
     * this will open mixer_fd and do initial setup */

    while ((ch = getopt(argc, argv, "m:r")) != EOF) {
	switch (ch) {
	case 'm':
	    mixer_device = malloc(strlen(optarg) + 1);
	    mixer_device = strdup(optarg);
	    break;
	case 'r':
	    ign = True;
	    break;
	}
    }
    if ((argc > 1) && !mixer_device && !ign)
	exit(-1);

    if (!mixer_device)
	mixer_device = "/dev/mixer";

    mixer_init(mixer_device);

    /* open display */
    dispname = getenv("DISPLAY");
    if (!(display = XOpenDisplay(dispname))) {
	eprint("Unable to open display '%s'\n", dispname);
	exit(-1);
    }

    screen = DefaultScreen(display);
    rootwin = DefaultRootWindow(display);
    depth = DefaultDepth(display, screen);
    maingc = XCreateGC(display, rootwin, 0, 0);

    xpma.exactColors = 0;
    xpma.alloc_close_colors = 1;
    xpma.closeness = 32767;
    xpma.valuemask = XpmExactColors | XpmAllocCloseColors | XpmCloseness;
    if (XpmCreatePixmapFromData
	(display, rootwin, main_xpm, &mainpanel, NULL,
	 &xpma) != XpmSuccess) {
	eprint("Cannot allocate colors for the main pixmap\n");
	exit(-1);
    }

    /* this is the "handle" of the "mixer" hardware */
    handle =
	XGetImage(display, mainpanel, 27, 139, 15, 11, AllPlanes, ZPixmap);
    onhandle =
	XGetImage(display, mainpanel, 27, 151, 15, 11, AllPlanes, ZPixmap);

    /* create background pixmap for the main mixer window */
    mixer = XCreatePixmap(display, rootwin, 226, 128, depth);
    XCopyArea(display, mainpanel, mixer, maingc, 0, 0, 226, 128, 0, 0);
    /* mark channels that don't exist with a red light and erase
     * "mixer" hardware. */
    make_caps();

    win = XCreateSimpleWindow(display, rootwin, 0, 0, 226, 116, 1, 0, 0);
    XSelectInput(display, win,
		 ButtonPressMask | ExposureMask | ButtonReleaseMask |
		 PointerMotionMask | StructureNotifyMask | KeyPressMask);
    sizehints.width = 226;
    sizehints.height = 116;
    sizehints.min_width = sizehints.width;
    sizehints.min_height = sizehints.height;
    sizehints.max_width = sizehints.width;
    sizehints.max_height = sizehints.height;
    sizehints.flags = USSize | PMinSize | PMaxSize;
    XSetWMNormalHints(display, win, &sizehints);

    xattr.save_under = True;
    xattr.backing_store = WhenMapped;
    XChangeWindowAttributes(display, win, CWSaveUnder | CWBackingStore,
			    &xattr);

    XStoreName(display, win, "Mixer");
    XSetIconName(display, win, "Mixer");

    XSetWindowBackgroundPixmap(display, win, mixer);
    XMapWindow(display, win);
    XFlush(display);

    if ((ret = fork()) > 0)
	exit(0);
    if (ret == -1)
	perror("forking");
    if (setsid() < 0)
	perror("detaching from tty");

    while (1) {
	/* check if the mixer has been updated or not */
	ioctl(mixer_fd, SOUND_MIXER_INFO, &mixer_info);

	if (mixer_info.modify_counter != modify_counter) {
	    for (i = 0; i < 8; i++) {
		if (mixer_data[i].cap) {
		    int foo;
		    ioctl(mixer_fd, MIXER_READ(mixer_data[i].type), &foo);
		    mixer_data[i].volume = ((foo >> 8) + (foo & 0xFF)) / 2;
		    if (mixer_data[i].volume)
			mixer_data[i].mute = False;
		    else
			mixer_data[i].mute = True;
		    draw_switches(mixer_data[i].volume, i);
		    draw_slider(i, mixer_data[i].volume);
		}
	    }
	    ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &recsource);
	    draw_switches(recsource, 16);
	    modify_counter = mixer_info.modify_counter;
	}

	while (XPending(display)) {
	    XNextEvent(display, &Event);
	    switch (Event.type) {
	    case Expose:
		/* loop to get rid of all expose events */
		while (XCheckTypedWindowEvent(display, win, Expose, &bob));
		for (j = 0; j < 8; j++) {
		    if (mixer_data[j].cap) {
			draw_switches(mixer_data[j].volume, j);
			draw_slider(j, mixer_data[j].volume);
		    }
		}
		draw_switches(recsource, 16);
		XFlush(display);
		break;
	    case DestroyNotify:
		XCloseDisplay(display);
		exit(0);
		break;
	    case ButtonPress:
		pressEvent(&Event.xbutton);
		break;
	    case ButtonRelease:
		releaseEvent(&Event.xbutton);
		break;
	    case KeyPress:
		keyEvent(Event.xkey.keycode, Event.xkey.state);
		break;
	    case MotionNotify:
		motionEvent(&Event.xmotion);
		break;
	    }
	}
	usleep(100000);
    }
    exit(0);
}

void releaseEvent(XButtonEvent * button)
{
    dragging = 0;
}

void draw_switches(int state, int type)
{
    XImage *temp;
    XImage *btn;

    btn =
	XGetImage(display, mainpanel, 27, 130, 31, 8, AllPlanes, ZPixmap);
    temp =
	XGetImage(display, mainpanel, 27, 118, 49, 11, AllPlanes, ZPixmap);

    if (type != 16) {
	if (state)
	    XPutImage(display, win, maingc, temp, 25, 0, 8 + (type * 27),
		      18, 24, 11);
	else
	    XPutImage(display, win, maingc, temp, 0, 0, 8 + (type * 27),
		      18, 24, 11);
    }

    /* with record source we assume we can always capture from line mic or cd */
    if (type == 16) {
	XPutImage(display, win, maingc, btn,
		  (state == (1 << SOUND_MIXER_LINE)) ? 16 : 0, 0, 7, 95,
		  15, 8);
	XPutImage(display, win, maingc, btn,
		  (state == (1 << SOUND_MIXER_MIC)) ? 16 : 0, 0, 42, 95,
		  15, 8);
	XPutImage(display, win, maingc, btn,
		  (state == (1 << SOUND_MIXER_CD)) ? 16 : 0, 0, 75, 95, 15,
		  8);
    }
}

int checkreg(x, x1, y, y1, posx, posy)
{
    if (posx >= x && posx <= x1 && posy >= y && posy <= y1)
	return 1;
    else
	return 0;
}

void set_record(void)
{
    /* this is not fatal since we really don't care */
    if (ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &recsource) == -1)
	printf("Error setting mixer recording source!\n");
}


void keyEvent(unsigned int keycode, unsigned int state)
{
    int oldslider;

    switch (keycode) {
    case 23:			/* 'tab' (next/prev channel) (also left/right) */
    case 100:
    case 102:
	if (state || keycode == 100) {	/* modifier pressed */
	    oldslider = currslider;
	    currslider--;
	    while (mixer_data[currslider].cap == False)
		currslider--;
	    if (currslider < 0)
		currslider = 7;
	    draw_slider(oldslider, mixer_data[oldslider].volume);
	    draw_slider(currslider, mixer_data[currslider].volume);
	} else {		/* normal tab */
	    oldslider = currslider;
	    currslider++;
	    while (mixer_data[currslider].cap == False)
		currslider++;
	    if (currslider > 7)
		currslider = 0;
	    draw_slider(oldslider, mixer_data[oldslider].volume);
	    draw_slider(currslider, mixer_data[currslider].volume);
	}
	break;
    case 24:			/* 'q' "Quit" */
	XCloseDisplay(display);
	exit(0);
	break;
    case 52:			/* 'z' "Line" */
	recsource = 1 << SOUND_MIXER_LINE;
	draw_switches(recsource, 16);
	set_record();
	break;
    case 53:			/* 'x' "Mic" */
	recsource = 1 << SOUND_MIXER_MIC;
	draw_switches(recsource, 16);
	set_record();
	break;
    case 54:			/* 'c' "CD" */
	recsource = 1 << SOUND_MIXER_CD;
	draw_switches(recsource, 16);
	set_record();
	break;
    case 58:			/* 'm' "Mute" */
	if (mixer_data[currslider].mute) {	/* was muted, restore */
	    mixer_data[currslider].volume =
		mixer_data[currslider].oldvolume;
	    mixer_data[currslider].mute = False;
	    draw_slider(currslider, mixer_data[currslider].volume);
	    draw_switches(1, currslider);
	} else {		/* not muted, save & mute */
	    mixer_data[currslider].oldvolume =
		mixer_data[currslider].volume;
	    mixer_data[currslider].mute = True;
	    mixer_data[currslider].volume = -2;
	    draw_slider(currslider, 0);
	    draw_switches(0, currslider);
	}
	setmix(currslider);
    case 98:			/* UP */
	mixer_data[currslider].volume += 2;
	if (mixer_data[currslider].volume < 0)
	    mixer_data[currslider].volume = 0;
	if (mixer_data[currslider].volume > 100)
	    mixer_data[currslider].volume = 100;
	draw_slider(currslider, mixer_data[currslider].volume);
	setmix(currslider);
	break;
    case 104:			/* DOWN */
	mixer_data[currslider].volume -= 2;
	if (mixer_data[currslider].volume < 0)
	    mixer_data[currslider].volume = 0;
	if (mixer_data[currslider].volume > 100)
	    mixer_data[currslider].volume = 100;
	draw_slider(currslider, mixer_data[currslider].volume);
	setmix(currslider);
	break;
    }
}

void pressEvent(XButtonEvent * button)
{
    int x = button->x;
    int y = button->y;
    int i;
    int oldslider;
    dragging = 1;

    if (checkreg(7, 22, 95, 103, x, y)) {
	recsource = 1 << SOUND_MIXER_LINE;
	draw_switches(recsource, 16);
	set_record();
    }
    if (checkreg(42, 57, 95, 103, x, y)) {
	recsource = 1 << SOUND_MIXER_MIC;
	draw_switches(recsource, 16);
	set_record();
    }
    if (checkreg(75, 90, 95, 103, x, y)) {
	recsource = 1 << SOUND_MIXER_CD;
	draw_switches(recsource, 16);
	set_record();
    }

    if (checkreg(114, 216, 88, 108, x, y))
	exit(0);

    for (i = 0; i <= 8; i++) {
	if (checkreg((i * 27) + 11, (i * 27) + 26, 36, 76, x, y)
	    && mixer_data[i].cap) {
	    oldslider = currslider;
	    currslider = 255;	/* bogus */
	    draw_slider(oldslider, mixer_data[oldslider].volume);
	    currslider = i;
	    mixer_data[i].volume = offset;
	    draw_slider(currslider, offset);
	    if (mixer_data[currslider].mute && offset) {
		draw_switches(y, currslider);
		mixer_data[currslider].mute = False;
	    }
	    if (offset == 0) {
		mixer_data[currslider].mute = True;
		draw_switches(0, currslider);
	    }
	    setmix(currslider);
	}

	/* mute */
	if (checkreg((i * 27) + 11, (i * 27) + 26, 20, 27, x, y)
	    && mixer_data[i].cap) {
	    if (mixer_data[i].mute) {	/* was muted, restore */
		mixer_data[i].volume = mixer_data[i].oldvolume;
		mixer_data[i].mute = False;
		draw_slider(i, mixer_data[i].volume);
		draw_switches(1, i);
	    } else {		/* not muted, save & mute */
		mixer_data[i].oldvolume = mixer_data[i].volume;
		mixer_data[i].mute = True;
		mixer_data[i].volume = 0;
		draw_slider(i, 0);
		draw_switches(0, i);
	    }
	    setmix(i);
	}
    }
}


void motionEvent(XMotionEvent * motion)
{
    int x = motion->x;
    int y = motion->y;

    if (x >= (currslider * 27) + 11 && x <= (currslider * 27) + 26
	&& y >= 36 && y <= 76 && dragging) {
	mixer_data[currslider].volume = offset;
	draw_slider(currslider, offset);
	if (mixer_data[currslider].mute && offset) {
	    draw_switches(y, currslider);
	    mixer_data[currslider].mute = False;
	}
	if (offset == 0) {
	    mixer_data[currslider].mute = True;
	    draw_switches(0, currslider);
	}
	setmix(currslider);
    }
}

/* draw slider 0 .. slider, using setting "vol" (0 .. 100) */
void draw_slider(int slider, int vol)
{
    XClearArea(display, win, (slider * 27) + 11, 29, 15, 51, 0);
    if (slider == currslider) {
	XPutImage(display, win, maingc, onhandle, 0, 0, (slider * 27) + 11,
		  (69 - ((vol * 40) / 100)), 15, 11);
    } else {
	XPutImage(display, win, maingc, handle, 0, 0, (slider * 27) + 11,
		  (69 - ((vol * 40) / 100)), 15, 11);
    }
}

void setmix(int channel)
{
    int foo;
    foo = (mixer_data[channel].volume << 8) + mixer_data[channel].volume;
    ioctl(mixer_fd, MIXER_WRITE(mixer_data[channel].type), &foo);
}
